import bcrypt


def to_utf8(ps):
    return ps.encode("utf-8")


def hash_password(password):
    return bcrypt.hashpw(to_utf8(password), bcrypt.gensalt()).decode("utf8")


def passwords_match(hashed_password, password):
    return bcrypt.checkpw(to_utf8(password), to_utf8(hashed_password))



import databases
import sqlalchemy
from fastapi.testclient import TestClient
from pytest import fixture
from sqlalchemy.orm import sessionmaker

from whitebox.core.settings import get_settings
from whitebox.entities.Base import Base
from whitebox.main import app
from whitebox.sdk.whitebox import Whitebox
from whitebox.tests.utils.maps import v1_test_order_map
from secrets import token_hex

settings = get_settings()


def get_order_number(task):
    return v1_test_order_map.index(task)


@fixture(scope="session")
def client():
    with TestClient(app) as client:
        yield client


@fixture(scope="session", autouse=True)
async def db():
    # runs once before all tests
    engine = sqlalchemy.create_engine(settings.POSTGRES_DB_URI)
    database = databases.Database(settings.POSTGRES_DB_URI)
    SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)
    Base.metadata.create_all(engine)
    await database.connect()
    db = SessionLocal()
    yield db
    # runs once after all tests
    await database.disconnect()
    # await database.execute(query="DROP DATABASE test WITH (FORCE);")
    Base.metadata.drop_all(engine)


class TestsState:
    user: dict = {}
    api_key: str = token_hex(32)
    model_binary: dict = {}
    model_multi: dict = {}
    model_multi_2: dict = {}
    model_multi_3: dict = {}
    inference_row_multi: dict = {}
    inference_row_binary: dict = {}


state = TestsState()


class TestsSDKState:
    wb: Whitebox


state_sdk = TestsSDKState()



from datetime import datetime
from whitebox.schemas.model import ModelType
from sklearn.datasets import load_breast_cancer, load_wine
import pandas as pd
import random
from copy import deepcopy

from whitebox.schemas.modelMonitor import AlertSeverity, MonitorMetrics, MonitorStatus

user_create_payload = dict(username="admin")

model_binary_create_payload = dict(
    name="Model 1",
    description="Model 1 description",
    type=ModelType.binary,
    features={
        "feature1": "numerical",
        "feature2": "numerical",
        "feature3": "numerical",
        "feature4": "numerical",
        "target": "numerical",
    },
    labels={"label_1": 0, "label_2": 1},
    prediction="target",
    probability="n/a",
)

model_multi_create_payload = dict(
    name="Model 2",
    description="Model 2 description",
    type=ModelType.multi_class,
    features={
        "feature1": "numerical",
        "feature2": "numerical",
        "feature3": "numerical",
        "feature4": "numerical",
        "target": "numerical",
    },
    labels={"label_1": 0, "label_2": 1, "label_3": 2},
    prediction="target",
    probability="n/a",
)

model_multi_2_create_payload = dict(
    name="Model 3",
    description="Model 3 description",
    type=ModelType.multi_class,
    features={
        "feature1": "numerical",
        "feature2": "numerical",
        "feature3": "numerical",
        "feature4": "numerical",
        "target": "numerical",
    },
    labels={"label_1": 0, "label_2": 1, "label_3": 2},
    prediction="target",
    probability="n/a",
)

model_multi_3_create_payload = dict(
    name="Model 4",
    description="Model 4 description",
    type=ModelType.multi_class,
    features={
        "feature1": "numerical",
        "feature2": "numerical",
        "feature3": "numerical",
        "feature4": "numerical",
        "target": "numerical",
    },
    labels={"label_1": 0, "label_2": 1, "label_3": 2},
    prediction="target",
    probability="n/a",
)

model_update_payload = dict(
    name="Model 1 - categorical",
    description="Model 1 description",
)

# dataset rows data for both binary and multiclass models
dataset_rows_single_row_column_payload = [
    {
        "nonprocessed": {},
        "processed": {"additionalProp1": 0, "additionalProp2": 0, "target": 1},
    }
]

dataset_rows_no_prediction_column_payload = [
    {
        "nonprocessed": {},
        "processed": {"additionalProp1": 0, "additionalProp2": 0},
    },
    {
        "nonprocessed": {},
        "processed": {"additionalProp1": 1, "additionalProp2": 0},
    },
]

dataset_rows_one_prediction_value_payload = [
    {
        "nonprocessed": {},
        "processed": {"additionalProp1": 0, "additionalProp2": 0, "target": 0},
    },
    {
        "nonprocessed": {},
        "processed": {"additionalProp1": 1, "additionalProp2": 0, "target": 0},
    },
]


df_load_multi = load_wine()
df_multi = pd.DataFrame(df_load_multi.data, columns=df_load_multi.feature_names)
df_multi["target"] = df_load_multi.target
df_multi = df_multi.tail(100)

dict_multi_data = df_multi.to_dict(orient="records")
dataset_rows_create_multi_class_payload = [
    {"processed": x, "nonprocessed": x} for x in dict_multi_data
]

df_load_binary = load_breast_cancer()
df_binary = pd.DataFrame(df_load_binary.data, columns=df_load_binary.feature_names)
df_binary["target"] = df_load_binary.target
df_binary = df_binary.tail(100)

dict_binary_data = df_binary.to_dict(orient="records")
dataset_rows_create_binary_payload = [
    {"processed": x, "nonprocessed": x} for x in dict_binary_data
]

dataset_rows_create_wrong_model_payload = list(
    (
        dict(
            model_id="wrong_model_id",
            nonprocessed={"sex": "male"},
            processed={"sex": 0},
        ),
        dict(
            model_id="wrong_model_id",
            nonprocessed={"sex": "female"},
            processed={"sex": 1},
        ),
    )
)


# inference rows data for both binary and multiclaas models
df_multi_inference = df_multi.tail(10)

dict_inferences = df_multi_inference.to_dict(orient="records")
inference_row_create_many_multi_payload = [
    {
        "timestamp": str(datetime.now()),
        "processed": x,
        "nonprocessed": x,
        "actual": random.randint(0, 1),
    }
    for x in dict_inferences
]

inference_row_create_single_row_payload = inference_row_create_many_multi_payload[0]
inference_row_create_many_multi_no_actual_payload = deepcopy(
    inference_row_create_many_multi_payload
)
for x in inference_row_create_many_multi_no_actual_payload:
    del x["actual"]

inference_row_create_many_multi_mixed_actuals_payload = (
    inference_row_create_many_multi_no_actual_payload
    + inference_row_create_many_multi_payload
)

# This is the body of the request coming from the sdk
df_binary_inference = df_binary.tail(10)

dict_inferences = df_binary_inference.to_dict(orient="records")
inference_row_create_many_binary_payload = [
    {
        "timestamp": str(datetime.now()),
        "processed": x,
        "nonprocessed": x,
        "actual": random.randint(0, 1),
    }
    for x in dict_inferences
]

timestamps = pd.Series(["2022-12-22T12:13:27.879738"] * 10)
mixed_actuals = pd.Series([0, 1, None, 1, 0, None, None, 1, 0, None])

model_monitor_accuracy_create_payload = dict(
    name="accuracy monitor ",
    status=MonitorStatus.active,
    metric=MonitorMetrics.accuracy,
    lower_threshold=0.85,
    severity=AlertSeverity.low,
    email="example@whitebox.io",
)

model_monitor_f1_create_payload = dict(
    name="f1 Monitor",
    status=MonitorStatus.active,
    metric=MonitorMetrics.f1,
    lower_threshold=0.85,
    severity=AlertSeverity.low,
    email="example@whitebox.io",
)

model_monitor_data_drift_create_payload = dict(
    name="data drift monitor",
    status=MonitorStatus.active,
    metric=MonitorMetrics.data_drift,
    feature="concavity error",
    severity=AlertSeverity.low,
    email="example@whitebox.io",
)

model_monitor_precision_create_payload = dict(
    name="precision monitor",
    status=MonitorStatus.active,
    metric=MonitorMetrics.precision,
    lower_threshold=0.85,
    severity=AlertSeverity.low,
    email="example@whitebox.io",
)



from whitebox.core.settings import get_settings
import databases
import sqlalchemy
from sqlalchemy.orm import sessionmaker
from whitebox.entities.Base import Base
from whitebox.schemas.user import UserCreateDto

from whitebox import crud
from whitebox.utils.passwords import hash_password
from whitebox.utils.logger import cronLogger as logger
import os

from secrets import token_hex

settings = get_settings()
database = databases.Database(settings.POSTGRES_DB_URI)
engine = sqlalchemy.create_engine(settings.POSTGRES_DB_URI)
SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)


def get_db():
    db = SessionLocal()
    try:
        yield db
    finally:
        db.close()


async def connect():
    """
    Connect to DB
    """
    Base.metadata.create_all(engine)
    db = SessionLocal()
    if not os.getenv("ENV") == "test":
        admin_exists = crud.users.get_first_by_filter(db=db, username="admin")
        if not admin_exists:
            api_key = token_hex(32)
            obj_in = UserCreateDto(username="admin", api_key=hash_password(api_key))
            crud.users.create(db=db, obj_in=obj_in)
            logger.info(f"Created username: admin, API key: {api_key}")
    await database.connect()


async def close():
    """
    Close DB Connection
    """
    await database.disconnect()
    # logging.info("Closed connection with DB")



from whitebox.schemas.user import User, UserCreateDto
from fastapi import APIRouter, Depends, status
from whitebox import crud
from sqlalchemy.orm import Session
from whitebox.core.db import get_db
from whitebox.schemas.utils import StatusCode
from whitebox.utils.passwords import hash_password
from whitebox.utils.errors import add_error_responses

users_router = APIRouter()


@users_router.post(
    "/users",
    tags=["Users"],
    response_model=User,
    summary="Create user",
    status_code=status.HTTP_201_CREATED,
    responses=add_error_responses([400, 409]),
)
async def create_user(body: UserCreateDto, db: Session = Depends(get_db)) -> User:
    """Creates an admin user during testing"""
    body.api_key = hash_password(body.api_key)
    new_user = crud.users.create(db=db, obj_in=body)
    return new_user


@users_router.delete(
    "/users/{user_id}",
    tags=["Users"],
    response_model=StatusCode,
    summary="Delete user",
    status_code=status.HTTP_200_OK,
    responses=add_error_responses([404]),
)
async def delete_user(
    user_id: str,
    db: Session = Depends(get_db),
) -> StatusCode:
    """Deletes the admin user during testing"""
    crud.users.remove(db=db, _id=user_id)
    return {"status_code": status.HTTP_200_OK}



from enum import Enum
import numpy as np
import pandas as pd
from whitebox.schemas.model import FeatureTypes, ModelCreateDto, ModelType
from typing import Dict
import requests
import logging
from fastapi import status


class APiVersion(str, Enum):
    v1 = "v1"


logger = logging.getLogger(__name__)


class Whitebox:
    def __init__(
        self,
        host: str,
        api_key: str,
        verbose: bool = False,
        api_version: APiVersion = APiVersion.v1,
    ):
        self.host = host
        self.api_key = api_key
        self.verbose = verbose
        self.api_version = api_version

    def create_model(
        self,
        name: str,
        type: ModelType,
        features: Dict[str, FeatureTypes],
        prediction: str,
        probability: str,
        labels: Dict[str, int],
        description: str = "",
    ):
        """
        Create a new model in Whitebox and define a type, and a schema for it.
        """
        new_model = ModelCreateDto(
            name=name,
            description=description,
            type=type,
            features=features,
            labels=labels,
            prediction=prediction,
            probability=probability,
        )
        result = requests.post(
            url=f"{self.host}/{self.api_version}/models",
            json=new_model.dict(),
            headers={"api-key": self.api_key},
        )

        logger.info(result.json())
        return result.json()

    def get_model(self, model_id: str):
        """
        Returns a model by its id. If the model does not exist, returns None.
        """
        result = requests.get(
            url=f"{self.host}/{self.api_version}/models/{model_id}",
            headers={"api-key": self.api_key},
        )
        if result.status_code == status.HTTP_404_NOT_FOUND:
            return None

        return result.json()

    def delete_model(self, model_id: str):
        """
        Deletes a model by its id. If any error occurs, returns False.
        """
        result = requests.delete(
            url=f"{self.host}/{self.api_version}/models/{model_id}",
            headers={"api-key": self.api_key},
        )

        if result.status_code == status.HTTP_200_OK:
            return True

        return False

    def log_training_dataset(
        self, model_id: str, non_processed: pd.DataFrame, processed: pd.DataFrame
    ):
        """
        Logs a training dataset for a model.

        Non processed is a dataframe with the raw data.
        Processed is a dataframe with the data after it has been processed and before it has entered the model.
        """
        self._check_processed_and_non_processed_length(processed, non_processed)
        non_processed_json = non_processed.to_dict(orient="records")
        processed_json = processed.to_dict(orient="records")

        dataset_rows = []
        for i in range(len(non_processed)):
            dataset_rows.append(
                dict(
                    model_id=model_id,
                    nonprocessed=non_processed_json[i],
                    processed=processed_json[i],
                )
            )

        result = requests.post(
            url=f"{self.host}/{self.api_version}/dataset-rows",
            headers={"api-key": self.api_key},
            json=dataset_rows,
        )
        if result.status_code == status.HTTP_201_CREATED:
            return True

        return False

    def log_inferences(
        self,
        model_id: str,
        non_processed: pd.DataFrame,
        processed: pd.DataFrame,
        timestamps: pd.Series,
        actuals: pd.Series = None,
    ):
        """
        Logs inferences of a model.

        Non processed is a dataframe with the raw data.
        Processed is a dataframe with the data after it has been processed and before it has entered the model.
        """
        self._check_processed_and_non_processed_length(processed, non_processed)
        non_processed_json = non_processed.to_dict(orient="records")
        processed_json = processed.to_dict(orient="records")
        timestamps_list = timestamps.tolist()

        if actuals is not None:
            actuals = actuals.replace({np.nan: None})
            actuals_list = actuals.tolist()
        else:
            actuals_list = actuals

        inference_rows = []
        for i in range(len(non_processed)):
            inference_rows.append(
                dict(
                    model_id=model_id,
                    nonprocessed=non_processed_json[i],
                    processed=processed_json[i],
                    timestamp=timestamps_list[i],
                    actual=actuals_list[i] if actuals_list is not None else None,
                )
            )

        result = requests.post(
            url=f"{self.host}/{self.api_version}/inference-rows/batch",
            headers={"api-key": self.api_key},
            json=inference_rows,
        )
        if result.status_code == status.HTTP_201_CREATED:
            return True

        return False

    def _check_processed_and_non_processed_length(
        self, processed: pd.DataFrame, non_processed: pd.DataFrame
    ) -> bool:
        """
        Checks if the processed and non processed dataframes have the same number of rows.
        """
        if len(processed) != len(non_processed):
            raise ValueError(
                "Processed and non processed dataframes must have the same length."
            )
        return True



from whitebox.tests.v1.mock_data import user_create_payload
import pytest
from whitebox import schemas
from whitebox.tests.v1.conftest import get_order_number, state
from fastapi import status


@pytest.mark.order(get_order_number("users_create"))
def test_user_create(client):
    response = client.post(
        "/v1/users",
        json={**user_create_payload, "api_key": state.api_key},
    )
    assert response.status_code == status.HTTP_201_CREATED
    state.user = response.json()
    assert response.json()["id"] is not None
    assert response.json()["username"] == user_create_payload["username"]
    validated = schemas.User(**response.json())


@pytest.mark.order(get_order_number("users_delete"))
def test_user_delete(client):
    response = client.delete(
        f"/v1/users/{state.user['id']}",
    )
    assert response.json() == {"status_code": str(status.HTTP_200_OK)}



from typing import Dict, Optional
from pydantic import BaseModel
from whitebox.schemas.base import ItemBase
import enum


class ModelType(str, enum.Enum):
    binary = "binary"
    multi_class = "multi_class"


class FeatureTypes(str, enum.Enum):
    categorical = "categorical"
    boolean = "boolean"
    string = "string"
    datetime = "datetime"
    numerical = "numerical"


class ModelBase(BaseModel):
    name: str
    description: str
    type: ModelType

    """Inference row metadata"""

    # TODO do we need this?
    features: Dict[str, FeatureTypes]
    labels: Dict[str, int]

    prediction: str
    probability: str


class Model(ModelBase, ItemBase):
    pass


class ModelCreateDto(ModelBase):
    pass


class ModelUpdateDto(BaseModel):
    name: Optional[str]
    description: Optional[str]
    type: Optional[ModelType]



import pytest
from whitebox import schemas
from whitebox.tests.v1.conftest import get_order_number, state
from fastapi import status


@pytest.mark.order(get_order_number("performance_metrics_get_model_all"))
def test_performance_metric_get_model_all(client):
    response_multi = client.get(
        f"/v1/performance-metrics?model_id={state.model_multi['id']}",
        headers={"api-key": state.api_key},
    )
    response_binary = client.get(
        f"/v1/performance-metrics?model_id={state.model_binary['id']}",
        headers={"api-key": state.api_key},
    )
    response_wrong_model = client.get(
        f"/v1/performance-metrics?model_id=wrong_model_id",
        headers={"api-key": state.api_key},
    )

    assert response_multi.status_code == status.HTTP_200_OK
    assert response_binary.status_code == status.HTTP_200_OK
    assert response_wrong_model.status_code == status.HTTP_404_NOT_FOUND

    validated = [schemas.MultiClassificationMetrics(**m) for m in response_multi.json()]
    validated = [
        schemas.BinaryClassificationMetrics(**m) for m in response_binary.json()
    ]



from sqlalchemy import Column, ForeignKey, String, DateTime, JSON
from whitebox.entities.Base import Base
from whitebox.utils.id_gen import generate_uuid


class DriftingMetric(Base):
    __tablename__ = "drifting_metrics"

    id = Column(String, primary_key=True, unique=True, default=generate_uuid)
    model_id = Column(String, ForeignKey("models.id", ondelete="CASCADE"))
    timestamp = Column(DateTime)
    concept_drift_summary = Column(JSON)
    data_drift_summary = Column(JSON)
    created_at = Column(DateTime)
    updated_at = Column(DateTime)
